﻿/*
 * SPiDER.Controls.js
 *
 * Copyright 2020
 * Carnegie Robotics LLC
 * 4501 Hatfield Street
 * Pittsburgh PA 15201
 * https://carnegierobotics.com
 */

SPiDER.controls = null;

SPiDER.Controls = function (object, domElement) {

    this.object = object;
    this.domElement = (domElement === undefined) ? document : domElement;
    this.enabled = true;
    this.viewMode = SPiDER.EXTERIOR_VIEW;
    this.phi = 0.0;
    this.rotation = 0.0;

    var self = this;
    var mousePosition = new THREE.Vector2();
    var zoom = [1.0, 1.0, 1.0, 1.0, 1.0];
    var objectQuaternion = null;

    this.reset = function () {

        zoom = [1.0, 1.0, 1.0, 1.0, 1.0];
        if (SPiDER.camera) {
            SPiDER.camera.zoom = 1.0;
            SPiDER.camera.updateProjectionMatrix();
        }

        object.quaternion.set(0.0, 0.0, 0.0, 1.0);
        object.rotateX(Math.PI / 2.0);
        objectQuaternion = null;
        self.phi = 0.0;
        self.rotation = 0.0;

        object.position.y = 1.6;
        if (recording.boundingBox) {
            object.position.y = 1.6 + Math.abs(recording.boundingBox.min.z - recording.boundingBox.max.z) / 2.0;
        }

        if (self.viewMode === SPiDER.INTERIOR_VIEW) {
            object.position.z = 0.0;
        } else {
            object.position.z = -2.5;
        }
    }

    this.setViewMode = function (viewMode) {

        if (self.viewMode === viewMode) {
            return;
        }

        if (viewMode === SPiDER.UNWRAPPED_VIEW) {

            objectQuaternion = self.object.quaternion;
            object.quaternion.set(0.0, 0.0, 0.0, 1.0);
            object.rotateX(Math.PI / 2.0);

            self.object.position.z = -2.5;

            recording.meshGroup.children.forEach(function (o) {
                o.children[0].material = o.children[0].unwrapMaterial;
                o.children[0].material.side = THREE.FrontSide;
                o.children[0].material.uniforms.rotation.value = self.rotation;
                o.children[0].material.needsUpdate = true;
            });

            if (SPiDER.camera) {
                var lookAt = new THREE.Vector3(0.0, 1.6, -1.0);
                SPiDER.camera.lookAt(lookAt);
                SPiDER.camera.updateProjectionMatrix();
            }

            self.viewMode = viewMode;

        }
        else if (viewMode === SPiDER.INTERIOR_VIEW) {

            if (objectQuaternion !== null) {
                object.quaternion = objectQuaternion;
                objectQuaternion = null;
            }

            self.object.position.z = 0.0;
            recording.meshGroup.children.forEach(function (o) {
                o.children[0].material = o.children[0].meshMaterial;
                o.children[0].material.side = THREE.FrontSide;
                o.children[0].material.needsUpdate = true;
            });

            if (SPiDER.camera) {
                var lookAt = new THREE.Vector3(0.0, 1.6 + Math.sin(self.phi), -Math.cos(self.phi));
                SPiDER.camera.lookAt(lookAt);
                SPiDER.camera.updateProjectionMatrix();
            }

            self.viewMode = viewMode;

        }
        else if (viewMode === SPiDER.EXTERIOR_VIEW) {

            if (objectQuaternion !== null) {
                object.quaternion = objectQuaternion;
                objectQuaternion = null;
            }

            self.object.position.z = -2.5;
            recording.meshGroup.children.forEach(function (o) {
                o.children[0].material = o.children[0].meshMaterial;
                o.children[0].material.side = THREE.BackSide;
                o.children[0].material.needsUpdate = true;
            });

            if (SPiDER.camera) {
                var lookAt = new THREE.Vector3(0.0, 1.6, -1.0);
                SPiDER.camera.lookAt(lookAt);
                SPiDER.camera.updateProjectionMatrix();
            }

            self.viewMode = viewMode;

        }

        if (SPiDER.camera) {
            SPiDER.camera.zoom = zoom[self.viewMode];
            SPiDER.camera.updateProjectionMatrix();
        }
    }

    this.handleGamepadInput = function (gamepad) {

        if (gamepad && gamepad.hand == 'left') {
            object.position.x += gamepad.axes[0] * 0.01;
            object.position.z += gamepad.axes[1] * 0.01;
        }
        if (gamepad && gamepad.hand == 'right') {
            object.position.y += gamepad.axes[1] * 0.01;
            object.rotateZ(-gamepad.axes[0] * 0.01);
        }

    }

    function onMouseWheel(e) {

        if (SPiDER.camera) {
            SPiDER.camera.zoom *= (1.0 - e.deltaY * 0.01);
            if (self.viewMode === SPiDER.INTERIOR_VIEW && SPiDER.camera.zoom < 1.0) {
                SPiDER.camera.zoom = 1.0;
            }
            zoom[self.viewMode] = SPiDER.camera.zoom;
            SPiDER.camera.updateProjectionMatrix();
        }

    }

    function onMouseDown(e) {
        mousePosition.x = e.screenX;
        mousePosition.y = e.screenY;
        self.domElement.addEventListener('mousemove', onMouseMove);
        document.addEventListener('mouseup', onMouseUp);
    }

    function onMouseMove(e) {

        var deltaX = (mousePosition.x - e.screenX) * 0.001;
        var deltaY = (mousePosition.y - e.screenY) * 0.001;
        mousePosition.x = e.screenX;
        mousePosition.y = e.screenY;

        if (self.viewMode === SPiDER.UNWRAPPED_VIEW) {
            self.rotation -= (deltaX * 2);
            if (self.rotation < -Math.PI) {
                self.rotation += 2.0 * Math.PI;
            }
            else if (self.rotation > Math.PI) {
                self.rotation -= 2.0 * Math.PI;
            }

            object.position.y += deltaY;
            if (object.position.y > recording.path[recording.path.length - 1][2] + 1.6) {
                object.position.y = recording.path[recording.path.length - 1][2] + 1.6;
            }
            else if (object.position.y < recording.path[0][2] + 1.6) {
                object.position.y = recording.path[0][2] + 1.6;
            }

            recording.meshGroup.children.forEach(function (o) {
                o.children[0].material.uniforms.rotation.value = self.rotation;
            });
        }
        else {
            if (self.viewMode === SPiDER.INTERIOR_VIEW) {
                deltaX = -deltaX;

                if ((e.buttons & 1) == 1) {
                    self.phi -= deltaY;
                    self.phi = Math.min(self.phi, Math.PI / 2 - 5 * Math.PI / 180.0);
                    self.phi = Math.max(self.phi, -Math.PI / 2 + 5 * Math.PI / 180.0);
                    var lookAt = new THREE.Vector3(0.0, 1.6 + Math.sin(self.phi), -Math.cos(self.phi));
                    SPiDER.camera.lookAt(lookAt);
                    SPiDER.camera.updateProjectionMatrix();
                }
            }

            if (self.viewMode === SPiDER.EXTERIOR_VIEW || (e.buttons & 2) == 2) {
                var offset = 0.0;
                if (self.viewMode === SPiDER.INTERIOR_VIEW) {
                    deltaY *= 3;
                    offset = 2.0;
                }
                object.position.y += deltaY;
                if (object.position.y > recording.path[recording.path.length - 1][2] + 1.6) {
                    object.position.y = recording.path[recording.path.length - 1][2] + 1.6;
                }
                else if (object.position.y < recording.path[0][2] + 1.6 - offset) {
                    object.position.y = recording.path[0][2] + 1.6 - offset;
                }

            }
            self.object.rotateZ(deltaX);
        }
    }

    function onMouseUp(e) {

        document.removeEventListener('mouseup', onMouseUp);
        self.domElement.removeEventListener('mousemove', onMouseMove);

    }

    function onTouchStart(e) {
    }

    function onTouchMove(e) {
    }

    function onTouchEnd(e) {
    }

    domElement.addEventListener('wheel', onMouseWheel);
    domElement.addEventListener('mousedown', onMouseDown);
    domElement.addEventListener('touchstart', onTouchStart);
    domElement.addEventListener('touchmove', onTouchMove);
    domElement.addEventListener('touchend', onTouchEnd);
};

SPiDER.Controls.prototype.constructor = SPiDER.Controls;
