﻿var scene;
var renderer;
//var camera;
//var controls;
var cameraNeedsReset = false;

var recording = {
    pointAnnotations: {},
    linearAnnotations: {},
    zeroReference: null,
    clockReference: null,
    models: new THREE.Group()
};

var measurementMode = 0;
var capturedPoints = [];

var UnwrapShader = {
    uniforms: {
        'rotation': { type: 'f', value: 0.0 },
        'texture1': { type: 't', value: null }
    },
    vertexShader: [
        'uniform float rotation;',
        'varying vec2 vUv;',
        'varying float visible;',

        '#define M_PI 3.1415926535897932384626433832795',

        'void main() {',
        '    vUv = uv;',
        '    vec4 vpos = vec4(position[0], position[1], position[2], 1.0);',

        '    float angle = atan (vpos[1], vpos[0]) + rotation;',
        '    if (angle < -M_PI) { angle = angle + 2.0 * M_PI; }',
        '    if (angle > M_PI) { angle = angle - 2.0 * M_PI; }',
        '    float magnitude = sqrt (vpos[0]*vpos[0] + vpos[1] * vpos[1]);',

        '    vpos[0] = angle;',
        '    vpos[1] = -magnitude;',
        '    gl_Position = projectionMatrix * modelViewMatrix * vpos;',

        '    visible = 1.0;',
        '    if (angle < -M_PI + 0.25 || angle > M_PI - 0.25) {',
        '        visible = 0.0;',
        '    }',

//        '    gl_Position = projectionMatrix * modelViewMatrix * vec4 (position, 1.0);',
        '}'
    ].join('\n'),
    fragmentShader: [
        'uniform sampler2D texture1;',
        'varying vec2 vUv;',
        'varying float visible;',

        'void main() {',
        '    if (visible < 0.5) {',
        '        discard;',
        '    } else {',
        '        gl_FragColor = texture2D (texture1, vUv);',
        '        gl_FragColor.a = visible;',
        '    }',
        '}'
    ].join('\n')
};

var loadingManager = new THREE.LoadingManager();

function loadModelMetadata(metadataName) {
    $.get(recording.baseUrl + '/' + metadataName, function (data) {
        if (metadataName == 'controls.json') {
            recording.path = data.segmented_path;
            recording.boundingBox = null;

            SPiDER.controls = new SPiDER.Controls(recording.models, renderer.domElement);
            SPiDER.controls.reset();
        }
    }, 'json');

    for (var i = 0; i < recording.meshResources.files.length; ++i) {
        loadMesh(recording.meshResources.files[i]);
    }

    for (var i = 0; i < recording.pointResources.files.length; ++i) {
        loadPointCloud(recording.pointResources.files[i]);
    }
}

function loadMesh(meshName) {
    var mtlLoader = new THREE.MTLLoader(loadingManager);
    mtlLoader.load(recording.baseUrl + '/' + meshName.replace(/\.[^\.]+$/, '.mtl'), function (materials) {
        materials.preload();

        var objLoader = new THREE.OBJLoader(loadingManager);
        objLoader.setMaterials(materials);
        objLoader.load(recording.baseUrl + '/' + meshName, function (object) {
            object.children[0].material.vertexColors = THREE.NoColors;
            object.children[0].material.needsUpdate = true;
            object.children[0].meshMaterial = object.children[0].material;
            object.children[0].unwrapMaterial = new THREE.ShaderMaterial({
                uniforms: {
                    'rotation': { type: 'f', value: 0.0 },
                    'texture1': { type: 't', value: object.children[0].material.map }
                },
                vertexShader: UnwrapShader.vertexShader,
                fragmentShader: UnwrapShader.fragmentShader
            });
            object.children[0].unwrapMaterial.map = object.children[0].material.map;
            if (SPiDER.controls) {
                if (SPiDER.controls.viewMode === SPiDER.EXTERIOR_VIEW) {
                    object.children[0].material.side = THREE.BackSide;
                } else if (SPiDER.controls.viewMode == SPiDER.UNWRAPPED_VIEW) {
                    object.children[0].material = object.children[0].unwrapMaterial;
                }
            }

            recording.meshGroup.add(object);

            if (!recording.boundingBox) {
                recording.boundingBox = new THREE.Box3().setFromObject(object);
                if (SPiDER.controls) {
                    SPiDER.controls.reset();
                }
            }
        });
    });
}

function loadPointCloud(name) {
    var loader = new PCLoader(loadingManager);
    loader.load(recording.baseUrl + '/' + name, function (object) {
        recording.pointCloud.add(object);

        if (!recording.boundingBox) {
            recording.boundingBox = new THREE.Box3().setFromObject(object);
            if (SPiDER.controls) {
                SPiDER.controls.reset();
            }
        }
    });
}

function loadRecording(baseUrl) {
    $.get(baseUrl, function(data) {

        recording.baseUrl = baseUrl;
        recording.files = data;
        recording.pointResources = { files: [] };
        recording.meshResources = { files: [] };

        scene.add(recording.models);

        recording.meshGroup = new THREE.Group();
        recording.meshGroup.visible = false;
        recording.models.add(recording.meshGroup);

        recording.pointCloud = new THREE.Group();
        recording.pointCloud.name = "recordingPointCloud";
        recording.models.add(recording.pointCloud);

        for (var i = 0; i < data.length; ++i) {
            if (/\.xyzrgb$/.test(data[i])) {
                recording.pointResources.files.push(data[i]);
            }
            if (/^color.+\.obj$/.test(data[i])) {
                recording.meshResources.files.push(data[i]);
            }
        }

        if (recording.files.includes('model.json')) {
            loadModelMetadata('model.json');
        }
        else if (recording.files.includes('controls.json')) {
            loadModelMetadata('controls.json');
        }

    }, 'json');
}

function resizeCanvas() {
    if (renderer) {
        var width = $(renderer.domElement).parent().width();
        var height = $(renderer.domElement).parent().height();

        if (SPiDER.camera) {
            if (SPiDER.camera.isOrthographicCamera) {
                var scaleFactor = 3.0 * Math.sin(Math.PI * (25.0 / 180.0)) / height;
                SPiDER.camera.left = -width * scaleFactor;
                SPiDER.camera.right = width * scaleFactor;
                SPiDER.camera.top = height * scaleFactor;
                SPiDER.camera.bottom = -height * scaleFactor;
            }
            else if (SPiDER.camera.isPerspectiveCamera) {
                SPiDER.camera.aspect = width / height;
            }
            SPiDER.camera.updateProjectionMatrix();
        }

        renderer.setSize(width, height, false);
    }
}

function nextDefaultColor(count) {
    h = (count * 65) % 360;
    x = Math.floor((1 * (1 - Math.abs((h / 60) % 2 - 1))) * 255);
    if (h < 60) {
        r = 255; g = x; b = 0;
    }
    else if (h < 120) {
        r = x; g = 255; b = 0;
    }
    else if (h < 180) {
        r = 0; g = 255; b = x;
    }
    else if (h < 240) {
        r = 0; g = x; b = 255;
    }
    else if (h < 300) {
        r = x; g = 0; b = 255;
    }
    else {
        r = 255; g = 0; b = x;
    }
    return '#' + ('00' + r.toString(16)).slice(-2) + ('00' + g.toString(16)).slice(-2) + ('00' + b.toString(16)).slice(-2)
}

function onDblClick(e) {

    var x = e.pageX - $(renderer.domElement).offset().left;
    var y = e.pageY - $(renderer.domElement).offset().top;

    var mousePosition = new THREE.Vector2(x / renderer.domElement.width * 2.0 - 1.0, -y / renderer.domElement.height * 2.0 + 1.0);
    var raycaster = new THREE.Raycaster();
    raycaster.setFromCamera(mousePosition, SPiDER.camera);

    var count = 0;
    var selectedPoint = new THREE.Vector3(0.0, 0.0, 0.0);
    if (recording.meshGroup.visible) {
        raycaster.params.Mesh.threshold = 0.1;
        recording.meshGroup.children.forEach(function (obj) {
            var intersections = raycaster.intersectObjects(obj.children);
            if (intersections.length > 0) {
                selectedPoint.add(intersections[0].point);
                count++;
            }
            console.log(count);
        });
    }
    if (recording.pointCloud.visible) {
        raycaster.params.Points.threshold = 0.02;
        recording.pointCloud.children.forEach(function (obj) {
            var intersections = raycaster.intersectObject(obj);
            if (intersections.length > 0) {
                selectedPoint.add(intersections[0].point);
                count++;
            }
        });
    }
    if (count > 0) {
        selectedPoint.divideScalar(count);

        var transform = new THREE.Matrix4();
        transform.getInverse(recording.models.matrix);
        selectedPoint.applyMatrix4(transform);

        capturedPoints.push(selectedPoint);
    }

    var color = nextDefaultColor(Object.keys(recording.pointAnnotations).length + Object.keys(recording.linearAnnotations).length);
    if (measurementMode == 1 && capturedPoints.length == 1) {
        var geometry = new THREE.SphereGeometry(0.02, 32, 32);
        var material = new THREE.MeshBasicMaterial({
            color: color,
            depthTest: false,
            opacity: 0.5,
            transparent: true
        });
        var object = new THREE.Mesh(geometry, material);
        object.userData.isPointAnnotation = true;
        object.position.set(selectedPoint.x, selectedPoint.y, selectedPoint.z);
        recording.models.add(object);

        var grid = $('#pointAnnotations_grid').data('kendoGrid');

        grid.dataSource.add({
            Name: 'Point #' + (grid.dataSource.total() + 1),
            Type: '',
            PositionX: selectedPoint.x,
            PositionY: selectedPoint.y,
            PositionZ: selectedPoint.z,
            Size: 0.02,
            Color: color.slice(-6),
            Notes: ''
        });
        grid.dataSource.sync();

        var data = grid.dataSource.data();
        recording.pointAnnotations[data[data.length - 1].uid] = object;
        renderer.domElement.removeEventListener('dblclick', onDblClick);
        capturedPoints.length = 0;
    }
    if (measurementMode == 2 && capturedPoints.length == 2) {
        var object = new THREE.Group();
        var geometry1 = new THREE.SphereGeometry(0.015, 32, 32);
        var geometry2 = new THREE.SphereGeometry(0.015, 32, 32);
        var material = new THREE.MeshBasicMaterial({
            color: color,
            depthTest: false,
            opacity: 0.5,
            transparent: true
        });
        object.userData.isLinearAnnotation = true;
        var object1 = new THREE.Mesh(geometry1, material);
        object1.position.set(capturedPoints[0].x, capturedPoints[0].y, capturedPoints[0].z);
        object.add(object1);
        var object2 = new THREE.Mesh(geometry2, material);
        object2.position.set(capturedPoints[1].x, capturedPoints[1].y, capturedPoints[1].z);
        object.add(object2);
        var geometry3 = new THREE.Geometry();
        geometry3.vertices.push(new THREE.Vector3(capturedPoints[0].x, capturedPoints[0].y, capturedPoints[0].z));
        geometry3.vertices.push(new THREE.Vector3(capturedPoints[1].x, capturedPoints[1].y, capturedPoints[1].z));
        object.add(new THREE.Line(geometry3, new THREE.LineBasicMaterial({ color: color, opacity: 0.5, transparent: true, depthTest: false })));
        recording.models.add(object);

        var grid = $('#linearAnnotations_grid').data('kendoGrid');
        grid.dataSource.add({
            Name: 'Measure #' + (grid.dataSource.total() + 1),
            Position1X: capturedPoints[0].x,
            Position1Y: capturedPoints[0].y,
            Position1Z: capturedPoints[0].z,
            Position2X: capturedPoints[1].x,
            Position2Y: capturedPoints[1].y,
            Position2Z: capturedPoints[1].z,
            Size: 0.02,
            Color: color.slice(-6),
            Notes: '',
            HLength: Math.sqrt(Math.pow(capturedPoints[1].x - capturedPoints[0].x, 2) + Math.pow(capturedPoints[1].y - capturedPoints[0].y, 2)),
            VLength: Math.abs(capturedPoints[1].z - capturedPoints[0].z),
            Length: Math.sqrt(Math.pow(capturedPoints[1].x - capturedPoints[0].x, 2) + Math.pow(capturedPoints[1].y - capturedPoints[0].y, 2) + Math.pow(capturedPoints[1].z - capturedPoints[0].z, 2))
        });
        grid.dataSource.sync();

        var data = grid.dataSource.data();
        recording.linearAnnotations[data[data.length - 1].uid] = object;
        renderer.domElement.removeEventListener('dblclick', onDblClick);
        capturedPoints.length = 0;
    }
}

$(function () {
    scene = new THREE.Scene();
    scene.add(new THREE.AmbientLight(0xFFFFFF, 0.005));
    scene.children[0].color.r = 128;
    scene.children[0].color.g = 128;
    scene.children[0].color.b = 128;

    renderer = new THREE.WebGLRenderer({
        canvas: document.getElementById('webgl_viewer'),
        antialias: true
    });
    renderer.setPixelRatio(window.devicePixelRatio);
    renderer.domElement.addEventListener('contextmenu', function (e) {
        e.preventDefault();
    });

    var vrButton = THREE.WEBVR.createButton(renderer, { referenceSpaceType: 'local' });
    $('#page_splitter .k-pane:last-child').append(vrButton);
    $(vrButton).on('click', function () {
        if (renderer.vr) {
            renderer.vr.enabled = !renderer.vr.enabled;
            if (!renderer.vr.enabled) {
                cameraNeedsReset = true;
            }
        }
    });

    SPiDER.createDefaultCamera(renderer.domElement);

    renderer.setAnimationLoop(function () {
        if (cameraNeedsReset) {
            resizeCanvas();
            cameraNeedsReset = false;
        }

        var gamepads = navigator.getGamepads();
        for (var i = 0; i < gamepads.length; ++i) {
            if (gamepads[i]) {
                SPiDER.controls.handleGamepadInput(gamepads[i]);
            }
        }

        renderer.render(scene, SPiDER.camera);
    });
});
